/******************************************************************************
 * FILE PURPOSE:  AES Encoder Key Expansion Routine 
 ******************************************************************************
 * FILE NAME:   aesKeyExpEnc.c  
 *
 * DESCRIPTION: AES Encryption Key Expansion Algorithm Implementation 
 *
 *              Note: It runs on MIPS/ARM or DSP
 *              So maintain this support while you modify this code.
 *
 * EXTERNAL COMPILE TIME FLAGS:
 *
 *              aes_ALL_KEY_SIZES:
 *                           If we want to support 192 and 256 bit keys
 *                           along with 128 bit key. (It will cost extra cycles!)
 *
 * The following source was taken from cryptographic software written by the below authors
 * and has been modified and adapted by TI
 *****************************************************************************
 * rijndael-alg-fst.c
 *
 * @version 3.0 (December 2000)
 *
 * Optimised ANSI C code for the Rijndael cipher (now AES)
 *
 * @author Vincent Rijmen <vincent.rijmen@esat.kuleuven.ac.be>
 * @author Antoon Bosselaers <antoon.bosselaers@esat.kuleuven.ac.be>
 * @author Paulo Barreto <paulo.barreto@terra.com.br>
 *
 * This code is hereby placed in the public domain.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/
 
/* AES header files */
#include "aes.h"

/* for 128-bit blocks, AES never uses more than 10 rcon values */
static const tulong rcon[] = 
{
	0x01000000, 0x02000000, 0x04000000, 0x08000000,
	0x10000000, 0x20000000, 0x40000000, 0x80000000,
	0x1B000000, 0x36000000 
};

extern const tulong aesTe4[];

/******************************************************************************
  * FUNCTION: aesKeyExpandEnc    
  * PURPOSE: Expand the cipher key into the encryption key schedule
  ******************************************************************************/
tint aesKeyExpandEnc(tulong *rk, tword cipherKey[], tint keyBits)
{
    tint i = 0;
    tulong temp;
    tulong *Te4 = (tulong *)&aesTe4;
  
    /* Input the cipher key as big Endian byte stream */
    secutil_GETU32_I(cipherKey, rk[0]);
    secutil_GETU32_I(cipherKey, rk[1]);
    secutil_GETU32_I(cipherKey, rk[2]);
    secutil_GETU32_I(cipherKey, rk[3]);
    
	if (keyBits == 128) {
		for (;;) {
			temp  = rk[3];
			rk[4] = rk[0] ^
				(Te4[(temp >> 16) & 0xff] & 0xff000000) ^
				(Te4[(temp >>  8) & 0xff] & 0x00ff0000) ^
				(Te4[(temp      ) & 0xff] & 0x0000ff00) ^
				(Te4[(temp >> 24)       ] & 0x000000ff) ^
				rcon[i];
			rk[5] = rk[1] ^ rk[4];
			rk[6] = rk[2] ^ rk[5];
			rk[7] = rk[3] ^ rk[6];
			if (++i == 10) {
				return 10;   /* Nr is fixed and set to 10 */
			}
			rk += 4;
		}
	}
    
#ifdef aes_ALL_KEY_SIZES    
    /* Cipher key expansion for keyBits > 128 */
	secutil_GETU32_I(cipherKey, rk[4]);
	secutil_GETU32_I(cipherKey, rk[5]);
	if (keyBits == 192) {
		for (;;) {
			temp = rk[ 5];
			rk[ 6] = rk[ 0] ^
				(Te4[(temp >> 16) & 0xff] & 0xff000000) ^
				(Te4[(temp >>  8) & 0xff] & 0x00ff0000) ^
				(Te4[(temp      ) & 0xff] & 0x0000ff00) ^
				(Te4[(temp >> 24)       ] & 0x000000ff) ^
				rcon[i];
			rk[ 7] = rk[ 1] ^ rk[ 6];
			rk[ 8] = rk[ 2] ^ rk[ 7];
			rk[ 9] = rk[ 3] ^ rk[ 8];
			if (++i == 8) {
				return 12;
			}
			rk[10] = rk[ 4] ^ rk[ 9];
			rk[11] = rk[ 5] ^ rk[10];
			rk += 6;
		}
	}

    /* Cipher key expansion for keyBits > 192 */
	secutil_GETU32_I(cipherKey, rk[6]);
	secutil_GETU32_I(cipherKey, rk[7]);
	if (keyBits == 256) {
        for (;;) {
        	temp = rk[ 7];
        	rk[ 8] = rk[ 0] ^
        		(Te4[(temp >> 16) & 0xff] & 0xff000000) ^
        		(Te4[(temp >>  8) & 0xff] & 0x00ff0000) ^
        		(Te4[(temp      ) & 0xff] & 0x0000ff00) ^
        		(Te4[(temp >> 24)       ] & 0x000000ff) ^
        		rcon[i];
        	rk[ 9] = rk[ 1] ^ rk[ 8];
        	rk[10] = rk[ 2] ^ rk[ 9];
        	rk[11] = rk[ 3] ^ rk[10];
			if (++i == 7) {
				return 14;
			}
        	temp = rk[11];
        	rk[12] = rk[ 4] ^
        		(Te4[(temp >> 24)       ] & 0xff000000) ^
        		(Te4[(temp >> 16) & 0xff] & 0x00ff0000) ^
        		(Te4[(temp >>  8) & 0xff] & 0x0000ff00) ^
        		(Te4[(temp      ) & 0xff] & 0x000000ff);
        	rk[13] = rk[ 5] ^ rk[12];
        	rk[14] = rk[ 6] ^ rk[13];
        	rk[15] = rk[ 7] ^ rk[14];

			rk += 8;
        }
	}

#endif  /* aes_ALL_KEY_SIZES */
	return 0;
}

/******************************************************************************
  * FUNCTION: aesInvKey    
  * PURPOSE: Derive the inverse key used at AES-CBC decryption operation
  ******************************************************************************/
void aesInvKey(tword invKey[], tword cipherKey[], tint keyBits)
{
    tint i = 0;
    tulong roundKey[60];
    tulong* rk = &roundKey[0];
#ifdef aes_ALL_KEY_SIZES    
    tulong temp;
    tulong *Te4 = (tulong *)&aesTe4;
#endif    
    tword *pInvKey = invKey;
    tulong *pInvRK = &roundKey[40];
    int keySize = keyBits >> 5;  /* Key size in 32-bit words */
    
	if (keyBits == 128) {
        aesKeyExpandEnc(rk, cipherKey, 128);
	}
    
#ifdef aes_ALL_KEY_SIZES    
    /* Cipher key expansion for keyBits > 128 */
	if (keyBits == 192) {
        /* Input the cipher key as big Endian byte stream */
        secutil_GETU32_I(cipherKey, rk[0]);
        secutil_GETU32_I(cipherKey, rk[1]);
        secutil_GETU32_I(cipherKey, rk[2]);
        secutil_GETU32_I(cipherKey, rk[3]);
	    secutil_GETU32_I(cipherKey, rk[4]);
	    secutil_GETU32_I(cipherKey, rk[5]);
		for (;;) {
			temp = rk[ 5];
			rk[ 6] = rk[ 0] ^
				(Te4[(temp >> 16) & 0xff] & 0xff000000) ^
				(Te4[(temp >>  8) & 0xff] & 0x00ff0000) ^
				(Te4[(temp      ) & 0xff] & 0x0000ff00) ^
				(Te4[(temp >> 24)       ] & 0x000000ff) ^
				rcon[i];
			rk[ 7] = rk[ 1] ^ rk[ 6];
			rk[ 8] = rk[ 2] ^ rk[ 7];
			rk[ 9] = rk[ 3] ^ rk[ 8];
			rk[10] = rk[ 4] ^ rk[ 9];
			rk[11] = rk[ 5] ^ rk[10];
			if (++i == 8) {
				break;
			}
			rk += 6;
		}
        pInvRK = &roundKey[48];
	}

    /* Cipher key expansion for keyBits > 192 */
	if (keyBits == 256) {
        aesKeyExpandEnc(rk, cipherKey, 256);
        pInvRK = &roundKey[52];
	}
#endif  /* aes_ALL_KEY_SIZES */


    /* Output the inverse key as big-endina byte stream */
    for (i = 0; i < keySize; i++)
    {
        secutil_PUTU32_I(pInvKey, pInvRK[i]); 
    }
}

/* nothing past this point */
