/* ========================================================================== */
/**
 *  @file   sha1.h
 *
 *  path    /dsps_gtmas/ti/mas/sha1/sha1.h
 *
 *  @brief  Header file of SHA1 module
 *
 *  ============================================================================
 *  
 * Copyright (C) 2011 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  ============================================================================
 */

/** @defgroup SHA1 */

#ifndef _SHA1_H
#define _SHA1_H

/* Required headers */
#include <ti/mas/types/types.h>
#include <ti/mas/secutil/secutil.h>

/** @ingroup SHA1 */
/** @{ */

/**
 *  @def sha1_NUM_WORDS_IN_BLOCK
 *  @brief Number of 32 bit words within the SHA1 operating block size of 512 bits
 */
#define sha1_NUM_WORDS_IN_BLOCK	                  (512/TYP_TULONG_SIZE)

/**
 *  @brief  Structure used by SHA1 algorithm to compute the message digest
 *
 */
typedef struct sha1Inst_s
{
	tulong h0;                            /**< H buffer 0 - used to compute first
	                                                                      32 bits of message digest. */
	tulong h1;                            /**< H buffer 1 - used to compute second
	                                                                      32 bits of message digest. */
	tulong h2;                            /**< H buffer 2 - used to compute third
	                                                                      32 bits of message digest. */
	tulong h3;                            /**< H buffer 3 - used to compute fourth
	                                                                      32 bits of message digest. */
	tulong h4;                            /**< H buffer 4 - used to compute fifth
	                                                                      32 bits of message digest. */
	tulong Nl;                             /**< Lower 32 bits of the number of bits 
	                                                    in the message to be hashed.  Up to 2^64 bits can 
	                                                    be hashed so two variables are used. */
	tulong Nh;                            /**< Upper 32 bits of the number of bits 
	                                                    in the message to be hashed.  Up to 2^64 bits can 
	                                                    be hashed so two variables are used. */	                                                    
	tulong data[sha1_NUM_WORDS_IN_BLOCK];  /**< The sixteen 32 bit words
	                                                                            of the message block being
	                                                                            operated on. */
	tint   num;                            /**< Holds the number of unhashed bytes to be hashed 
	                                                    by the sha1Final function, or a subsequent call to the
	                                                    sha1Update function, if the message is not a 
	                                                    multiple of 512. */
} sha1Inst_t;

/**
 *  @def sha1_SAVEKPADHASH
 *  @brief  Macro for copying the SHA1 algorithm instance of a completed hash of a block to a
 *             SHA1 instance storage entity.  This macro can be utilized at the HMAC level to
 *             to implement an optimization described in RFC2104 - Section 4.
 *
 */
#define sha1_SAVEKPADHASH(sha1Inst,k_pad_hash) memcpy(k_pad_hash,sha1Inst,sizeof(sha1Inst_t)) 


/* ============ sha1Init() ========== */
/**
 *  @brief      Initializes the sha1Inst algorithm structure.
 *
 *  @param[in, out]  inst     Pointer to an input unitialized algorithm instance.  Outputs a pointer
 *                                     to an initialized algorithm instance.
 *
 *  @param[in]  preInst      Precomputed, algorithm instance pointer.
 *                                     This input parameter allows the SHA1 code to conform to an
 *                                     HMAC level optimization described in RFC2104 - Section 4.
 *                                     Passing a NULL pointer through this parameter will initialize
 *                                     the SHA1 algorithm instance normally.  Passing a precomputed 
 *                                     sha1Inst_t through this parameter will cause the precomputed 
 *                                     SHA1 data to be copied to the currently in use SHA1 algorithm
 *                                     instance.                           
 *
 */
void sha1Init(void *inst, void *preInst);

/* ============ sha1Update() ========== */
/**
 *  @brief      Computes the hash of the input message up to the last block of data
 *                 which is greater than or equal to 512 bits long.  If there is leftover message
 *                 data it is returned through the algorithm instance pointer, as well as the 
 *                 computed message digest up to that point.
 *
 *  @param[in, out]  inst      Pointer to an input initialized algorithm instance.  Outputs a pointer
 *                                      to an algorithm instance containing remaining message data and
 *                                      the message digest computed up to that point.
 *
 *  @param[in]  data      Message data to be hashed.  Packed byte for 16 bit tword targets
 *
 *  @param[in]  len      Length of message to be hashed in bytes.
 *
 */
void sha1Update(void *inst, tword *data, tulong len);

/* ============ sha1Final() ========== */
/**
 *  @brief      Function where the final portion of the message which was less than 512 bits
 *                 is padded in accordance with the FIPS 180-1 standard, and hashed.
 *
 *  @param[in]  inst      Pointer to an input algorithm instance pointer containing remaining 
 *                                message and message digest.  Outputs a pointer to an algorithm
 *                                instance containing a completed message digest.
 *
 *  @param[out]  digest      Pointer to the 160-bit completed message digest.
 *
 */
void sha1Final(void *inst, tword *digest);

/** @} */ /** ingroup */

#endif /* _SHA1_H */
