/******************************************************************************
 * FILE PURPOSE:  SHA1 Message Authentication Algorithm
 ******************************************************************************
 * FILE NAME:   sha1.c  
 *
 * DESCRIPTION: Secure Hash Algorithm1 (SHA1) as per NIST FIPS PUB 180-1
 *
 *              Note: It runs on MIPS/ARM or DSP
 *              So maintain this support while you modify this code.
 *              
 * (C) Copyright 2008 Texas Instruments Inc - http://www.ti.com
 *
 * The following source was taken from cryptographic software written by Eric Young and has
 * been modified and adapted by TI
 *****************************************************************************
 * Copyright (C) 1995-1998 Eric Young (eay@cryptsoft.com)
 * All rights reserved.
 *
 * This package is an SSL implementation written
 * by Eric Young (eay@cryptsoft.com).
 * The implementation was written so as to conform with Netscapes SSL.
 * 
 * This library is free for commercial and non-commercial use as long as
 * the following conditions are aheared to.  The following conditions
 * apply to all code found in this distribution, be it the RC4, RSA,
 * lhash, DES, etc., code; not just the SSL code.  The SSL documentation
 * included with this distribution is covered by the same copyright terms
 * except that the holder is Tim Hudson (tjh@cryptsoft.com).
 * 
 * Copyright remains Eric Young's, and as such any Copyright notices in
 * the code are not to be removed.
 * If this package is used in a product, Eric Young should be given attribution
 * as the author of the parts of the library used.
 * This can be in the form of a textual message at program startup or
 * in documentation (online or textual) provided with the package.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    "This product includes cryptographic software written by
 *     Eric Young (eay@cryptsoft.com)"
 *    The word 'cryptographic' can be left out if the rouines from the library
 *    being used are not cryptographic related :-).
 * 4. If you include any Windows specific code (or a derivative thereof) from 
 *    the apps directory (application code) you must include an acknowledgement:
 *    "This product includes software written by Tim Hudson (tjh@cryptsoft.com)"
 * 
 * THIS SOFTWARE IS PROVIDED BY ERIC YOUNG ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The licence and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution licence
 * [including the GNU Public Licence.]
 *****************************************************************************/

/* Standard C headers */
#include <string.h>

#include <ti/mas/sha1/src/sha1loc.h>

/******************************************************************************
 * FUNCTION PURPOSE: SHA1 Instance Initialization
 ******************************************************************************
 * DESCRIPTION: Initializes the sha1 context
 *
 *    void sha1Init(         
 *      void            *inst,      - Pointer to an unitialized SHA1 instance.
 *      void            *preInst) - Precomputed, algorithm instance pointer.
 *                                     This input parameter allows the SHA1 code to conform to an
 *                                     HMAC level optimization described in RFC2104 - Section 4.
 *                                     Passing a NULL pointer through this parameter will initialize
 *                                     the SHA1 algorithm instance normally.  Passing a precomputed 
 *                                     SHA1inst through this parameter will cause the precomputed 
 *                                     SHA1 data to be copied to the currently in use SHA1 algorithm
 *                                     instance.      
 *****************************************************************************/
void sha1Init(void *inst, void *preInst)
{
  sha1Inst_t   *sha1Inst = (sha1Inst_t*)inst;
  sha1Inst_t   *k_pad_inst = (sha1Inst_t*)preInst;
  tuint i;

  /* If NULL is passed as the pointer to preInst set the sha1Inst to the initial
    * hashing parameters */
  if (!k_pad_inst)
  {
    sha1Inst->h0=sha1_INIT_DATA_h0;
    sha1Inst->h1=sha1_INIT_DATA_h1;
    sha1Inst->h2=sha1_INIT_DATA_h2;
    sha1Inst->h3=sha1_INIT_DATA_h3;
    sha1Inst->h4=sha1_INIT_DATA_h4;
    sha1Inst->Nl=0;
    sha1Inst->Nh=0;
    sha1Inst->num=0;
  } 
  /* If a pointer to a saved hash has been passed through the preInst point, set
    * the sha1Inst to the saved hash's parameters.  This is utilized in HMAC 
    * optimizations.  */
  else
  {
    sha1Inst->h0 = k_pad_inst->h0;
    sha1Inst->h1 = k_pad_inst->h1;
    sha1Inst->h2 = k_pad_inst->h2;
    sha1Inst->h3 = k_pad_inst->h3;
    sha1Inst->h4 = k_pad_inst->h4;
    sha1Inst->Nl = k_pad_inst->Nl;;
    sha1Inst->Nh = k_pad_inst->Nh;
    sha1Inst->num = k_pad_inst->num;

    for (i=0;i < sha1_NUM_WORDS_IN_BLOCK;i++)
    {
      sha1Inst->data[i] = k_pad_inst->data[i];
    }
  }   
}

/* nothing past this point */
